from .view import *

def PUI(func):
    """
    PUI.PUI triggers update() directly by redraw()
    """
    def func_wrapper(*args):
        class PUIViewWrapper(PUIView):
            def __init__(self, name):
                self.name = name
                super().__init__()

            def content(self):
                return func(*args)

        ret = PUIViewWrapper(func.__name__)
        return ret

    return func_wrapper


from .. import *
from .base import *

class MdiArea(QtBaseWidget):
    terminal = False
    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = QtWidgets.QMdiArea()

        super().update(prev)

    def addChild(self, idx, child):
        self.ui.addSubWindow(child.outer)

    def removeChild(self, idx, child):
        self.ui.removeSubWindow(child.outer)

    def addSubWindow(self, child):
        self.ui.addSubWindow(child.outer)

    def removeSubWindow(self, child):
        self.ui.removeSubWindow(child.outer)

class MdiSubWindow(QtBaseFrame):
    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = QtWidgets.QMdiSubWindow()

        super().update(prev)



from .. import *
from .base import *

class Text(TBase):
    def __init__(self, text, selectable=False):
        super().__init__()
        self.text = text

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.ui.update(self.text)
        else:
            self.ui = widgets.Label(self.text, markup=False)
        super().update(prev)

class Html(Text):
    supported = False

class MarkDown(TBase):
    weak_expand_x = True
    def __init__(self, text):
        super().__init__()
        self.text = text

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.ui.update(self.text)
        else:
            self.ui = widgets.Markdown(self.text)
        super().update(prev)



from .. import *
import tkinter as tk
from tkinter import ttk
from tkinter import font as tkFont
import functools

class TPUIView(PUIView):
    def redraw(self):
        if self.ui:
            self.ui.after(0, self.update)
        else:
            self.update()

class TkBaseWidget(PUINode):
    use_ttk = False
    def __init__(self, layout=None, side=None):
        super().__init__()
        self.layout_type = layout
        self.side = side

    @property
    def tkparent(self):
        parent = self.parent
        while not isinstance(parent, TkBaseWidget):
            parent = parent.parent
            if parent==parent.parent:
                return None
        return parent

    def destroy(self, direct):
        if self.ui:
            self.ui.destroy() # tk's destroy
            self.ui = None

    def update(self, prev):
        if not self.layout_width is None:
            self.ui.configure(width=self.layout_width)
        if not self.layout_height is None:
            self.ui.configure(height=self.layout_height)

        if self.use_ttk:
            styleKey = f"S{id(self.ui)}.{self.use_ttk}"
            tkstyle = ttk.Style()
            if not self.style_color is None:
                tkstyle.configure(styleKey, foreground=f"#{self.style_color:06X}")
            if not self.style_bgcolor is None:
                tkstyle.configure(styleKey, background=f"#{self.style_bgcolor:06X}")
            if self.style_fontfamily or self.style_fontsize:
                default = tkFont.nametofont('TkDefaultFont').actual()
                font_family = self.style_fontfamily or default["family"]
                font_size = self.style_fontsize or default["size"]
                font_weight = self.style_fontweight or default["weight"]
                tkstyle.configure(styleKey, font=(font_family, font_size, font_weight))
            self.ui.configure(style=styleKey)
        else:
            if not self.style_color is None:
                self.ui.configure(fg=f"#{self.style_color:06X}")
            if not self.style_bgcolor is None:
                self.ui.configure(bg=f"#{self.style_bgcolor:06X}")
            if self.style_fontfamily or self.style_fontsize:
                default = tkFont.nametofont('TkDefaultFont').actual()
                font_family = self.style_fontfamily or default["family"]
                font_size = self.style_fontsize or default["size"]
                font_weight = self.style_fontweight or default["weight"]
                self.ui.configure(font=(font_family, font_size, font_weight))
        super().update(prev)


# dprint = print
dprint = lambda *x: x

def recur_delete(node, child, direct):
    child.destroyed = True
    for sc in child.children:
        recur_delete(child, sc, False)
    child.destroy(direct)
    dprint("recur_delete", node.key, child, direct)

def sortDOM(dom):
    sorted = []
    ooo = []
    for e in dom:
        if e.outoforder:
            ooo.append(e)
        else:
            sorted.append(e)
    sorted.extend(ooo)
    return sorted

def sortGridDOM(dom):
    dom = [c for c in dom if c.grid_row is not None and c.grid_column is not None]
    return sorted(dom, key=lambda c:(c.grid_row, c.grid_column, c.grid_rowspan, c.grid_columnspan))

def sync(node, oldDOM, newDOM):
    if node.grid_layout:
        oldDOM = sortGridDOM(oldDOM)
        newDOM = sortGridDOM(newDOM)
    oldDOM = sortDOM(oldDOM)
    newDOM = sortDOM(newDOM)
    dprint("syncing", node.key, len(oldDOM), len(newDOM))

    dprint("  ===OLD===")
    for c in oldDOM:
        dprint("   ", c.key)

    dprint("  ===NEW===")
    for c in newDOM:
        dprint("   ", c.key)
    oldMap = [x.key for x in oldDOM]
    newMap = [x.key for x in newDOM]

    node.preSync()

    tbd = []
    for i,new in enumerate(newDOM):
        while True:
            # Step 1. skip common prefix
            if i < len(oldDOM) and oldMap[i] == new.key: # matched
                dprint(f"MATCHED {i} {new.key}")
                old = oldDOM[i]

                try:
                    new.update(old)
                except:
                    import traceback
                    print("## <ERROR OF update() >")
                    print(node.key)
                    traceback.print_exc()
                    print("## </ERROR OF update()>")

                if not new.terminal:
                    sync(new, old.children, new.children)

                break # finish

            # Step 2. trim removed nodes after common prefix
            trimmed = False
            while i < len(oldDOM) and not oldDOM[i].key in newMap: # trim old nodes
                dprint(f"TRIM {i} {oldDOM[i].key}")
                old = oldDOM.pop(i)
                oldMap.pop(i)
                node.removeChild(i, old)
                tbd.append(old)
                trimmed = True

            if trimmed:
                continue # restart

            # Step 3. setup target node
            try:
                idx = oldMap[i+1:].index(new.key)+i+1
            except ValueError:
                idx = None

            # Step 3-1. new node
            if idx is None:
                try:
                    new.update(None)
                except:
                    import traceback
                    print("## <ERROR OF update() >")
                    print(new.key)
                    traceback.print_exc()
                    print("## </ERROR OF update()>")
                if not new.terminal:
                    sync(new, [], new.children)
                node.addChild(i, new)
                oldDOM.insert(i, None) # placeholder
                oldMap.insert(i, new.key)

            # Step 3-2. existed node
            else:
                # if the target node is in just next position, requeue the front to prevent repositioning every nodes coming after
                # eg. when an element is removed from a long list, do single 3-2-1 instead of many 3-2-2

                # Step 3-2-1. yield the next position for the target node
                if idx==i+1:
                    oldMap.pop(i)
                    old = oldDOM.pop(i)
                    node.removeChild(i, old)

                    node.addChild(len(oldDOM), old)
                    oldMap.append(old.key)
                    oldDOM.append(old)

                    # sync will be peformed in next step 1

                    continue # restart

                 # Step 3-2-2. move target node
                else:
                    oldMap.pop(idx)
                    old = oldDOM.pop(idx)
                    node.removeChild(idx, old)

                    try:
                        new.update(old)
                    except:
                        import traceback
                        print("## <ERROR OF update() >")
                        print(node.key)
                        traceback.print_exc()
                        print("## </ERROR OF update()>")

                    if not new.terminal:
                        sync(new, old.children, new.children)

                    node.addChild(i, new)
                    oldDOM.insert(i, None) # placeholder
                    oldMap.insert(i, new.key)
            break # finish

    # Step 4. trim removed trail
    nl = len(newDOM)
    while len(oldDOM) > nl:
        old = oldDOM.pop(nl)
        oldMap.pop(nl)
        node.removeChild(nl, old)
        tbd.append(old)

    for c in newDOM:
        c.postUpdate()

    node.postSync()

    # release deleted nodes
    for old in tbd:
        recur_delete(node, old, True)



from .. import *
from .base import *

class TextField(TBase):
    content_width = None
    def __init__(self, model):
        super().__init__()
        self.model = model
        self.ui_text = self.model.value

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            model_value = self.model.value
            if self.ui_text != model_value:
                self.ui_text = model_value
                self.ui.value = model_value
        else:
            self.ui_text = self.model.value
            self.ui = widgets.Input(self.ui_text)

        self.ui.puinode = self
        super().update(prev)


    def _changed(self, value):
        self.ui_text = value
        self.model.value = value

    def _submitted(self, value):
        pass



from .. import *
from .base import *
from ..utils import *
from PySide6.QtWidgets import QSizePolicy

class QText(QtWidgets.QLabel):
    def __init__(self, text, sizeHint):
        self._sizehint = sizeHint
        super().__init__(text)

    def sizeHint(self):
        return QtCore.QSize(*self._sizehint)

    def minimumSizeHint(self):
        return QtCore.QSize(*self._sizehint)
    
    def setSizeHint(self, sizeHint):
        if sizeHint != self._sizehint:
            self._sizehint = sizeHint
            self.adjustSize()

class Text(QtBaseWidget):
    textformat = QtCore.Qt.TextFormat.PlainText
    def __init__(self, text, selectable=False, sizeHint=(120,320)):
        super().__init__()
        self.text = str(text)
        self.selectable = selectable
        self.sizeHint = sizeHint

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.ui.setText(self.text)
            self.ui.setSizeHint(self.sizeHint)
        else:
            self.ui = QText(self.text, self.sizeHint)
            self.ui.setTextFormat(self.textformat)
            self.ui.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)
            self.ui.setWordWrap(True)
            self.ui.setSizePolicy(QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Ignored))

        if self.selectable:
            self.ui.setTextInteractionFlags(QtCore.Qt.TextInteractionFlag.TextSelectableByMouse)
        else:
            self.ui.setTextInteractionFlags(QtCore.Qt.TextInteractionFlag.NoTextInteraction)

        super().update(prev)


class Html(Text):
    textformat = QtCore.Qt.TextFormat.RichText

class MarkDown(Text):
    textformat = QtCore.Qt.TextFormat.MarkdownText



from .. import *
from .base import *

class TextField(WxBaseWidget):
    def __init__(self, model):
        super().__init__()
        self.model = model
        self.editing = False
        self.changed_cb = None

    def update(self, prev):
        value = self.model.value
        if prev and prev.ui:
            self.editing = prev.editing
            self.ui = prev.ui
            if prev.last_value != value:
                if not self.editing or not value:
                    self.ui.SetValue(str(value))
            self.last_value = value
        else:
            self.last_value = value
            self.ui = wx.TextCtrl(getWindow(self.parent))
            self.ui.SetValue(str(value))
            self.ui.Bind(wx.EVT_TEXT, self.on_textchanged)
        super().update(prev)

    def change(self, cb, *args, **kwargs):
        self.changed_cb = (cb, args, kwargs)

    def on_textchanged(self, *args):
        node = self.get_node()
        node.editing = True
        node.model.value = self.ui.GetValue()
        if node.changed_cb:
            node.changed_cb[0](*node.changed_cb[1], **node.changed_cb[2])



from .. import *
from .base import *

class ProgressBar(FBase):
    def __init__(self, progress, maximum=100):
        super().__init__()
        self.progress = progress
        self.maximum = maximum

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.ui.value = self.progress / self.maximum
        else:
            self.ui = ft.ProgressBar(width=300) # XXX
            self.ui.value = self.progress / self.maximum
        self.ui.expand = self.layout_weight
        try:
            self.ui.update()
        except:
            pass
        super().update(prev)



from .. import *
from .base import *

class Button(FBase):
    def __init__(self, text):
        super().__init__()
        self.text = text

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.ui.text = self.text
            self.ui.on_click = self._clicked
            try:
                self.ui.update()
            except:
                pass
        else:
            self.ui = ft.ElevatedButton(text=self.text, on_click=self._clicked)
        self.ui.expand = self.layout_weight



from .. import *
from .base import *

class Tabs(QtBaseWidget):
    NORTH = "n"
    SOUTH = "s"
    EAST = "e"
    WEST = "w"
    terminal = False
    def __init__(self, tabposition=NORTH):
        super().__init__()
        self.tabposition = tabposition

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = QtWidgets.QTabWidget()
        super().update(prev)

    def addChild(self, idx, child):
        if not isinstance(child, Tab):
            raise RuntimeError("Tabs can only contain Tab")

        self._addChild(idx, child.children[0], child.label)

    def _addChild(self, idx, child, label):
        if isinstance(child, QtBaseWidget) or isinstance(child, QtBaseLayout):
            self.ui.insertTab(idx, child.outer, label)
        elif child.children:
            self._addChild(idx, child.children[0], label)

    def removeChild(self, idx, child):
        self.ui.removeTab(idx)


class Tab(PUINode):
    def __init__(self, label):
        super().__init__()
        self.label = label

    def addChild(self, idx, child):
        if idx > 0:
            raise RuntimeError("Tab can only have one child")



from .. import *
import sys

class Window(PUINode):
    def __init__(self, title=None, size=None, maximize=None, fullscreen=None):
        super().__init__()
        self.title = title
        self.size = size
        self.curr_size = None
        self.maximize = maximize
        self.curr_maximize = None
        self.fullscreen = fullscreen
        self.curr_fullscreen = None



def trbl(d):
    # https://www.w3schools.com/css/css_padding.asp
    if isinstance(d, int) or isinstance(d, float):
        return (d,d,d,d)
    elif len(d) == 4: # top, right, bottom, left
        return d
    elif len(d) == 3: # top, right/left, bottom
        return (d[0], d[1], d[2], d[1])
    elif len(d) == 2: # top/bottom, right/left
        return (d[0], d[1], d[0], d[1])
    else:
        raise RuntimeError(f"Unsupported format: {d}")
    
def trbl2ltrb(d):
    return (d[3], d[0], d[1], d[2])


from .. import *
from .base import *

class Scroll(PUINode):
    def __init__(self, vertical=None, horizontal=False):
        self.vertical = vertical
        self.horizontal = horizontal
        super().__init__()



from .. import *
from .base import *


def OpenDirectory(title="Open Directory", dir=None):
    res = QtWidgets.QFileDialog.getExistingDirectory(None, title, dir)
    return res

def OpenFile(title="Open File", dir=None, types=None):
    res = QtWidgets.QFileDialog.getOpenFileName(None, title, dir, types)
    return res[0] or None

def OpenFiles(title="Open Files", dir=None, types=None):
    res = QtWidgets.QFileDialog.getOpenFileNames(None, title, dir, types)
    return res[0] or None

def SaveFile(default, title="Save File", dir=None, types=None):
    if not dir:
        value = default
        if isinstance(value, str):
            dir = value
    res = QtWidgets.QFileDialog.getSaveFileName(None, title, dir, types)
    return res[0] or None

def Information(message="Information", title="Information Dialog"):
    dlg = QtWidgets.QMessageBox(None)
    dlg.setWindowTitle(title)
    dlg.setText(message)
    dlg.setStandardButtons(QtWidgets.QMessageBox.Ok)
    dlg.setIcon(QtWidgets.QMessageBox.Information)
    dlg.exec_()

def Warning(message="Warning", title="Warning Dialog"):
    dlg = QtWidgets.QMessageBox(None)
    dlg.setWindowTitle(title)
    dlg.setText(message)
    dlg.setStandardButtons(QtWidgets.QMessageBox.Ok)
    dlg.setIcon(QtWidgets.QMessageBox.Warning)
    dlg.exec_()

def Critical(message="Critical", title="Critical Dialog"):
    dlg = QtWidgets.QMessageBox(None)
    dlg.setWindowTitle(title)
    dlg.setText(message)
    dlg.setStandardButtons(QtWidgets.QMessageBox.Ok)
    dlg.setIcon(QtWidgets.QMessageBox.Critical)
    dlg.exec_()

def Confirm(message="Confirm", title="Confirm Dialog"):
    dlg = QtWidgets.QMessageBox(None)
    dlg.setWindowTitle(title)
    dlg.setText(message)
    dlg.setStandardButtons(QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
    dlg.setIcon(QtWidgets.QMessageBox.Question)
    button = dlg.exec_()

    if button == QtWidgets.QMessageBox.Yes:
        return True
    else:
        return False

def Prompt(prompt="Input", title="Input Dialog", default=""):
    text, ok = QtWidgets.QInputDialog.getText(None, title, prompt, QtWidgets.QLineEdit.Normal, default)
    if ok:
        return text
    else:
        return None



from .. import *
from .base import *

class RadioButton(TkBaseWidget):
    def __init__(self, text, value, model):
        super().__init__()
        self.var = None
        self.text = text
        self.value = value
        self.model = model

    def update(self, prev):
        if prev and prev.ui:
            self.var = prev.var
            self.ui = prev.ui
        else:
            self.var = tk.StringVar()
            self.ui = tk.Radiobutton(self.tkparent.inner, text=self.text, variable=self.var, value=self.value, anchor="w")

        self.var.set(self.model.value)
        self.ui.configure(command=self._select)
        super().update(prev)

    def _select(self):
        node = self.get_node()
        self.model.value = node.value



from .. import *
from .base import *

class Application(WxPUIView):
    def __init__(self):
        super().__init__()
        self.ui = None

    def update(self, prev=None):
        if not self.ui:
            self.ui = wx.App()

        super().update(prev)

    def addChild(self, idx, child):
        child.outer.Show()

    def removeChild(self, idx, child):
        child.outer.Close()

    def start(self):
        self.ui.MainLoop()



from .application import *
from .button import *
from .canvas import *
from .checkbox import *
from .label import *
from .layout import *
from .progressbar import *
from .radiobutton import *
from .scroll import *
from .tab import *
from .text import *
from .textfield import *
from .window import *
from .. import NotImplementedNode

Combobox = NotImplementedNode
ComboboxItem = NotImplementedNode

PUI_BACKEND = "flet"



from .. import *

import wx

def getWindow(p):
    from .window import Window
    while p:
        if isinstance(p, Window):
            return p.ui
        p = p.parent
    return None

class WxPUIView(PUIView):
    def __init__(self):
        super().__init__()

    def destroy(self, direct):
        if self.ui: # PUIView doesn't have ui
            self.ui.Destroy()
        self.ui = None
        super().destroy(direct)

    def update(self, prev=None):
        if self.retired_by:
            return
        self.dirty = False
        super().update(prev)
        self.updating = False
        if self.dirty:
            self.update(prev)

class WxBaseWidget(PUINode):
    terminal = True

    def __init__(self):
        super().__init__()

    def destroy(self, direct):
        self.ui.Destroy()
        self.ui = None
        super().destroy(direct)

class WxBaseLayout(PUINode):
    def __init__(self):
        super().__init__()

    def destroy(self, direct):
        # self.ui.Destroy()
        self.ui = None
        super().destroy(direct)

    def addChild(self, idx, child):
        from .layout import Spacer
        if isinstance(child, WxBaseLayout):
            self.ui.Insert(idx, child.outer)
        elif isinstance(child, WxBaseWidget):
            self.ui.Insert(idx, child.outer)
        elif isinstance(child, Spacer):
            self.ui.InsertStretchSpacer(idx, child.layout_weight or 1)
        elif child.children:
            self.addChild(idx, child.children[0])

    def removeChild(self, idx, child):
        from .layout import Spacer
        if isinstance(child, WxBaseLayout):
            self.ui.Detach(idx)
        elif isinstance(child, WxBaseWidget):
            self.ui.Detach(idx)
        elif isinstance(child, Spacer):
            self.ui.Detach(idx)
        else:
            self.removeChild(idx, child.children[0])



from .node import *
from .dom import *
from .state import StateMutationInViewBuilderError
import time

# dprint = print
dprint = lambda *x: x

class PUIView(PUINode):
    terminal = True
    __ALLVIEWS__  = []

    @staticmethod
    def reload():
        for v in PUIView.__ALLVIEWS__:
            v.redraw()

    def __init__(self, *args):
        self.frames = []
        self.dirty = False
        self.updating = False
        super().__init__(*args)
        PUIView.__ALLVIEWS__.append(self)

    def __enter__(self):
        if not hasattr(tls, "puistack"):
            tls.puistack = []
        tls.puistack.append(self)
        # print("enter", type(self).__name__, id(self))
        self.root.frames.append(self)
        return self

    def __exit__(self, ex_type, value, traceback):
        tls.puistack.pop()
        # print("exit", type(self).__name__, id(self))
        self.root.frames.pop()
        if ex_type is None: # don't consume exception
            return self

    def content(self):
        return None

    def dump(self):
        dprint(f"content() start", self.key)
        start = time.time()
        with self as scope:
            self.content()
        dprint(f"content() time: {time.time()-start:.5f}", self.key)
        return scope

    def destroy(self, direct):
        try:
            PUIView.__ALLVIEWS__.remove(self)
        except:
            pass
        return super().destroy(direct)

    def redraw(self):
        self.update()

    def update(self, prev=None):
        if self.retired_by:
            return
        if self.destroyed:
            return
        if not prev:
            if self.setup:
                self.setup()
                self.setup = None
        else:
            self.setup = None
        update_start = time.time()
        dprint("update()", self.key)
        if prev:
            self.children = prev.children
            prev.retired_by = self
            try:
                PUIView.__ALLVIEWS__.remove(prev)
            except:
                pass

        last_children = self.children
        self.children = []
        try:
            dprint(f"content() start", self.key)
            start = time.time()
            with self as scope: # init frame stack
                self.content() # V-DOM builder
            dprint(f"content() time: {time.time()-start:.5f}", self.key)
        except StateMutationInViewBuilderError:
            raise
        except:
            # prevent crash in hot-reloading
            self.children = last_children
            import traceback
            print("## <ERROR OF content() >", self.key, id(self))
            traceback.print_exc()
            print("## </ERROR OF content()>")
            return

        # print("PUIView.update", self) # print DOM

        start = time.time()
        dprint("sync() start", self.key)
        sync(self, last_children, self.children)
        dprint(f"sync() time: {time.time()-start:.5f}", self.key)

        dprint(f"update() time: {time.time()-update_start:.5f}", self.key)

    def setup(self):
        pass

    def run(self):
        self.redraw()
        self.start()



from .. import *
from .base import *
import math

class ScrollableFrame(ttk.Frame):
    def __init__(self, parent):
        self.vertical = None
        self.horizontal = False

        self.container = ttk.Frame(parent)
        self.container.rowconfigure(0, weight=1)
        self.container.columnconfigure(0, weight=1)

        self.scroll_y = tk.Scrollbar(self.container, orient=tk.VERTICAL)
        self.scroll_x = tk.Scrollbar(self.container, orient=tk.HORIZONTAL)
        self.should_scroll_y = False
        self.should_scroll_x = False
        self.last_bbox = None
        self.last_scroll_x = None
        self.last_scroll_y = None

        self.canvas = tk.Canvas(self.container)
        self.canvas.grid(row=0, column=0, sticky="nsew")
        super().__init__(self.canvas)
        self.canvas.create_window(0, 0, window=self, anchor="nw")

        self.bind("<Configure>", self._on_configure)

    def set_scroll(self, vertical, horizontal):
        self.vertical = vertical
        self.horizontal = horizontal

        self.update_scroller()

    def grid(self, *args, **kwargs):
        self.container.grid(*args, **kwargs)

    def configure(self, *args, **kwargs):
        self.container.configure(*args, **kwargs)

    def _on_configure(self, event):
        self.update_scroller()

    def update_scroller(self):
        bbox = self.canvas.bbox("all")
        _, _, w, h = bbox
        self.canvas.configure(scrollregion=bbox)
        if self.horizontal is False:
            self.canvas.configure(width=w)
        if self.vertical is False:
            self.canvas.configure(height=h)

        self.should_scroll_y = self.vertical or (self.vertical is None and h > self.canvas.winfo_height())
        if self.should_scroll_y:
            self.scroll_y.config(command=self.canvas.yview)
            self.canvas.config(yscrollcommand=self._on_scroll_y)
            self.scroll_y.grid(row=0, column=1, sticky="ns")

            self.canvas.bind_all("<MouseWheel>", self._on_mousewheel_y)
            self.canvas.bind("<Button-4>", self._on_mousewheel_y)
            self.canvas.bind("<Button-5>", self._on_mousewheel_y)
        else:
            self.scroll_y.grid_forget()

        self.should_scroll_x = self.horizontal or (self.horizontal is None and w > self.canvas.winfo_width())
        if self.should_scroll_x:
            self.scroll_x.config(command=self.canvas.xview)
            self.canvas.config(xscrollcommand=self._on_scroll_x)
            self.scroll_x.grid(row=1, column=0, sticky="ew")
        else:
            self.scroll_x.grid_forget()

    def _on_scroll_x(self, first, last):
        self.scroll_x.set(first, last)
        bbox = self.canvas.bbox("all")
        if self.last_bbox == bbox and self.last_scroll_x != (first,last):
            if float(first) < 0.05:
                self.puinode.align_x = 0
            elif float(last) > 0.95:
                self.puinode.align_x = 1
        else:
            self.last_bbox = bbox
            self.last_scroll_x = (first,last)

    def _on_scroll_y(self, first, last):
        self.scroll_y.set(first, last)
        bbox = self.canvas.bbox("all")
        if self.last_bbox == bbox and self.last_scroll_y != (first,last):
            if float(first) < 0.05:
                self.puinode.align_y = 0
            elif float(last) > 0.95:
                self.puinode.align_y = 1
        else:
            self.last_bbox = bbox
            self.last_scroll_y = (first,last)

    def _on_mousewheel_y(self, event):
        if not self.should_scroll_y:
            return
        if event.num == 5:
            delta = 1
        elif event.num == 4:
            delta = -1
        elif event.delta > 0:
            delta = 1
        elif event.delta < 0:
            delta = -1
        self.canvas.yview_scroll(delta, "units")

class Scroll(TkBaseWidget):
    use_ttk = "TFrame"
    END = -0.0
    def __init__(self, vertical=None, horizontal=False):
        self.vertical = vertical
        self.horizontal = horizontal
        self.align_x = 0
        self.align_y = 0
        super().__init__()
        self.layout_weight = 1

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.align_x = prev.align_x
            self.align_y = prev.align_y
        else:
            self.ui = ScrollableFrame(self.tkparent.inner)
        self.ui.puinode = self
        self.ui.set_scroll(self.vertical, self.horizontal)

        super().update(prev)

    def preSync(self):
        bbox = self.ui.canvas.bbox("all")
        _, _, w, h = bbox
        hsb = self.ui.scroll_x
        if self.align_x == 0:
            self.hsb_offset = hsb.get()[0] * w
        else:
            self.hsb_offset = w - hsb.get()[1] * w
        vsb = self.ui.scroll_y
        if self.align_y == 0:
            self.vsb_offset = vsb.get()[0] * h
        else:
            self.vsb_offset = h - vsb.get()[1] * h

    def postSync(self):
        self.ui.winfo_toplevel().update_idletasks()

        oldincx = self.ui.canvas["xscrollincrement"]
        oldincy = self.ui.canvas["yscrollincrement"]
        self.ui.canvas["xscrollincrement"] = 1
        self.ui.canvas["yscrollincrement"] = 1

        bbox = self.ui.canvas.bbox("all")
        _, _, w, h = bbox
        if self.align_x == 0:
            self.ui.canvas.xview_moveto(0.0)
            self.ui.canvas.xview_scroll(int(self.hsb_offset), "units")
        else:
            self.ui.canvas.xview_moveto(1.0)
            self.ui.canvas.xview_scroll(int(-self.hsb_offset), "units")
        if self.align_y == 0:
            self.ui.canvas.yview_moveto(0.0)
            self.ui.canvas.yview_scroll(int(self.vsb_offset), "units")
        else:
            self.ui.canvas.yview_moveto(1.0)
            self.ui.canvas.yview_scroll(int(-self.vsb_offset), "units")

        self.ui.canvas["xscrollincrement"] = oldincx
        self.ui.canvas["yscrollincrement"] = oldincy

    def addChild(self, idx, child):
        if idx:
            return
        child_outer = child.outer
        if child_outer:
            child_outer.grid(row=idx, column=0, sticky='nsew')

    def removeChild(self, idx, child):
        child.outer.grid_forget()

    def scrollX(self, pos=0):
        if math.copysign(1, pos) >= 0:
            self.align_x = 0
            self.hsb_offset = pos
        else:
            self.align_x = 1
            self.hsb_offset = abs(pos)
        return self

    def scrollY(self, pos=0):
        if math.copysign(1, pos) >= 0:
            self.align_y = 0
            self.vsb_offset = pos
        else:
            self.align_y = 1
            self.vsb_offset = abs(pos)
        return self


from .. import *
from .base import *
from PySide6.QtWidgets import QSizePolicy
import math

class Scroll(QtBaseWidget):
    terminal = False
    END = -0.0

    def __init__(self, vertical=None, horizontal=False):
        self.vertical = vertical
        self.horizontal = horizontal
        self.align_x = 0
        self.align_y = 0
        super().__init__()

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.align_x = prev.align_x
            self.align_y = prev.align_y
            if prev.vsb_conn:
                vsb = self.ui.verticalScrollBar()
                vsb.valueChanged.disconnect(prev.vsb_conn)
                vsb.rangeChanged.disconnect(prev.vsb_range_conn)
            if prev.hsb_conn:
                hsb = self.ui.horizontalScrollBar()
                hsb.valueChanged.disconnect(prev.hsb_conn)
                hsb.rangeChanged.disconnect(prev.hsb_range_conn)
        else:
            self.ui = QtWidgets.QScrollArea()
            self.ui.setWidgetResizable(True)
            if self.vertical is None:
                self.ui.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAsNeeded)
            elif self.vertical:
                self.ui.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOn)
            else:
                self.ui.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
            if self.horizontal is None:
                self.ui.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAsNeeded)
            elif self.horizontal:
                self.ui.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOn)
            else:
                self.ui.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        vsb = self.ui.verticalScrollBar()
        self.vsb_conn = vsb.valueChanged.connect(self.vsb_changed)
        self.vsb_range_conn = vsb.rangeChanged.connect(self.vsb_range_changed)
        hsb = self.ui.horizontalScrollBar()
        self.hsb_conn = hsb.valueChanged.connect(self.hsb_changed)
        self.hsb_range_conn = hsb.rangeChanged.connect(self.hsb_range_changed)
        super().update(prev)

    def addChild(self, idx, child):
        if idx != 0:
            return
        if isinstance(child, QtBaseWidget) or isinstance(child, QtBaseLayout):
            self.ui.setWidget(child.outer)
            if not hasattr(child.outer, "origResizeEvent"):
                child.outer.origResizeEvent = child.outer.resizeEvent
                child.outer.setSizePolicy(QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred))
            child.outer.resizeEvent = self.onUiResized
        elif child.children:
            self.addChild(idx, child.children[0])

    def removeChild(self, idx, child):
        if idx != 0:
            return
        if isinstance(child, QtBaseWidget) or isinstance(child, QtBaseLayout):
            child.outer.setParent(None)
        elif child.children:
            self.removeChild(idx, child.children[0])

    def onUiResized(self, event):
        node = self.get_node()
        if node.destroyed:
            return
        node.children[0].outer.origResizeEvent(event)
        node.children[0].outer.resizeEvent = self.onUiResized
        if node.horizontal is False:
            if isinstance(node.children[0], QtBaseLayout):
                node.outer.setMinimumWidth(node.children[0].outer.sizeHint().width())
            elif isinstance(node.children[0], QtBaseWidget):
                node.outer.setMinimumWidth(node.children[0].outer.sizeHint().width())

        if node.vertical is False:
            if isinstance(node.children[0], QtBaseLayout):
                node.outer.setMinimumHeight(node.children[0].outer.sizeHint().height())
            elif isinstance(node.children[0], QtBaseWidget):
                node.outer.setMinimumHeight(node.children[0].outer.sizeHint().height())

    def scrollX(self, pos=0):
        if math.copysign(1, pos) >= 0:
            self.align_x = 0
            self.hsb_offset = pos
        else:
            self.align_x = 1
            self.hsb_offset = abs(pos)
        return self

    def scrollY(self, pos=0):
        if math.copysign(1, pos) >= 0:
            self.align_y = 0
            self.vsb_offset = pos
        else:
            self.align_y = 1
            self.vsb_offset = abs(pos)
        return self

    def hsb_changed(self, *args, **kwargs):
        hsb = self.ui.horizontalScrollBar()
        v = hsb.value()
        if v < 10:
            self.align_x = 0
        elif v > hsb.maximum() - 10:
            self.align_x = 1

    def vsb_changed(self, *args, **kwargs):
        vsb = self.ui.verticalScrollBar()
        v = vsb.value()
        if v < 10:
            self.align_y = 0
        elif v > vsb.maximum() - 10:
            self.align_y = 1

    def preSync(self):
        hsb = self.ui.horizontalScrollBar()
        if self.align_x == 0:
            self.hsb_offset = hsb.value()
        else:
            self.hsb_offset = hsb.maximum() - hsb.value()
        vsb = self.ui.verticalScrollBar()
        if self.align_y == 0:
            self.vsb_offset = vsb.value()
        else:
            self.vsb_offset = vsb.maximum() - vsb.value()

    def vsb_range_changed(self, min, max):
        vsb = self.ui.verticalScrollBar()
        if self.align_y == 0:
            vsb.setValue(self.vsb_offset)
        else:
            vsb.setValue(max - self.vsb_offset)

    def hsb_range_changed(self, min, max):
        hsb = self.ui.verticalScrollBar()
        if self.align_y == 0:
            hsb.setValue(self.hsb_offset)
        else:
            hsb.setValue(max - self.hsb_offset)



from .. import *
from .base import *

class Tabs(TkBaseWidget):
    NORTH = "n"
    SOUTH = "s"
    EAST = "e"
    WEST = "w"
    terminal = False
    def __init__(self, tabposition=NORTH):
        super().__init__()
        self.tabposition = tabposition

    def destroy(self, direct):
        if direct:
            for w in self.widgets:
                if w:
                    w.deleteLater()
        super().destroy(direct)

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.childrenui = prev.childrenui
        else:
            self.ui = ttk.Notebook(self.tkparent.inner)
            self.childrenui = []
        super().update(prev)

    def addChild(self, idx, child):
        if not isinstance(child, Tab):
            raise RuntimeError("Tabs can only contain Tab")

        self._addChild(idx, child.children[0], child.label)

    def _addChild(self, idx, child, label):
        if isinstance(child, TkBaseWidget):
            if idx < len(self.childrenui):
                self.ui.insert(self.childrenui[idx], child.outer, text=label)
            else:
                self.ui.add(child.outer, text=label)
        elif child.children:
            self._addChild(idx, child.children[0], label)

    def removeChild(self, idx, child):
        if isinstance(child, TkBaseWidget):
            self.ui.forget(child.outer)
            self.childrenui.pop(idx)
        elif child.children:
            self.removeChild(idx, child.children[0])

class Tab(PUINode):
    def __init__(self, label):
        super().__init__()
        self.label = label

    def addChild(self, idx, child):
        if idx > 0:
            raise RuntimeError("TkNotebookFrame can only have one child")



__version__ = "0.3.2"

from .node import *
from .view import *
from .state import *
from .timeline import *
from .decorator import *

class NotImplementedNode():
    def __init__(self, *args, **kwargs):
        print("Not Implemented")
        import traceback
        import inspect
        traceback.print_stack(inspect.currentframe().f_back, 1)

    def __enter__(self):
        return self

    def __exit__(self, *args):
        pass

    def layout(self, *args, **kwargs):
        return self

    def style(self, *args, **kwargs):
        return self


from .. import *
from .base import *

class Window(WxBaseWidget):
    terminal = False

    def __init__(self, title=None, size=None, maximize=None, fullscreen=None):
        super().__init__()
        self.title = title
        self.size = size
        self.curr_size = None
        self.maximize = maximize
        self.curr_maximize = None
        self.fullscreen = fullscreen
        self.curr_fullscreen = None

    def update(self, prev=None):
        if prev and prev.ui:
            self.ui = prev.ui
            self.curr_size = prev.curr_size
            self.curr_maximize = prev.curr_maximize
            self.curr_fullscreen = prev.curr_fullscreen
        else:
            self.ui = wx.Frame(None)

        if self.curr_size != self.size:
            self.curr_size = self.size
            # self.ui.resize(*self.size)
        if self.curr_maximize !=  self.maximize:
            self.curr_maximize = self.maximize
            # self.ui.showMaximized()
        if self.curr_fullscreen != self.fullscreen:
            self.curr_fullscreen = self.fullscreen
            # self.ui.showFullScreen()
        if not self.title is None:
            pass
        super().update(prev)

    def addChild(self, idx, child):
        if isinstance(child, WxBaseLayout):
            self.ui.SetSizer(child.outer)
        elif isinstance(child, WxBaseWidget):
            pass
        else:
            self.addChild(idx, child.children[0])

    def removeChild(self, idx, child):
        if isinstance(child, WxBaseLayout):
            pass
        elif isinstance(child, WxBaseWidget):
            pass
        else:
            self.removeChild(idx, child.children[0])

    def postSync(self):
        self.ui.Layout()



from .. import *
from .base import *

class ToolBar(PUINode):
    outoforder = True
    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.actions = prev.actions
        else:
            self.ui = QtWidgets.QToolBar()
            self.actions = []

        super().update(prev)

    def addChild(self, idx, child):
        if idx < len(self.actions):
            if isinstance(child, ToolBarAction):
                self.actions.insert(idx, self.ui.insertAction(self.actions[idx], child.outer))
            else:
                self.actions.insert(idx, self.ui.insertWidget(self.actions[idx], child.outer))
        else:
            if isinstance(child, ToolBarAction):
                self.actions.append(self.ui.addAction(child.outer))
            else:
                self.actions.append(self.ui.addWidget(child.outer))

    def removeChild(self, idx, child):
        self.ui.removeAction(self.actions[idx])

class ToolBarAction(PUINode):
    def __init__(self, text, icon=None):
        super().__init__()
        self.text = text
        self.icon = icon
        self.onTriggered = None

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.ui.setText(self.text)
            self.ui.triggered.disconnect()
        else:
            self.ui = QtGui.QAction(self.text)
        if self.icon:
            self.ui.setIcon(self.icon)
        self.ui.triggered.connect(self._triggered)

        super().update(prev)

    def _triggered(self):
        node = self.get_node()
        if node.onTriggered:
            node.onTriggered[0](*node.onTriggered[1], **node.onTriggered[2])

    def trigger(self, callback, *cb_args, **cb_kwargs):
        self.onTriggered = (callback, cb_args, cb_kwargs)



from .. import *
from .base import *

class TextField(QtBaseWidget):
    def __init__(self, model):
        super().__init__()
        self.model = model
        self.editing = False
        self.changed_cb = None

    def update(self, prev):
        value = self.model.value
        if prev and prev.ui:
            self.editing = prev.editing
            self.ui = prev.ui
            self.ui.focusOutEvent = self.focusOutEvent
            if prev.last_value != value:
                self.ui.textChanged.disconnect()
                if not self.editing or not value:
                    self.ui.setText(str(value))
                self.ui.textChanged.connect(self.on_textchanged)
            self.last_value = value
        else:
            self.last_value = value
            self.ui = QtWidgets.QLineEdit()
            self.ui.origFocusOutEvent = self.ui.focusOutEvent
            self.ui.focusOutEvent = self.focusOutEvent
            self.ui.setText(str(value))
            self.ui.textChanged.connect(self.on_textchanged)
        super().update(prev)

    def change(self, cb, *args, **kwargs):
        self.changed_cb = (cb, args, kwargs)

    def on_textchanged(self):
        node = self.get_node()
        node.editing = True
        node.model.value = self.ui.text()
        if node.changed_cb:
            node.changed_cb[0](*node.changed_cb[1], **node.changed_cb[2])

    def focusOutEvent(self, event):
        node = self.get_node()
        node.editing = False
        node.ui.origFocusOutEvent(event)



from .. import *
from .base import *

from PySide6 import QtWidgets, QtGui
from PySide6.QtGui import QPainter, QColor
from PySide6.QtCore import QPoint

class PUIQtCanvas(QtWidgets.QWidget):
    def __init__(self, node, width=None, height=None):
        self.node = node
        self.width = width
        self.height = height
        super().__init__()

    def minimumSizeHint(self):
        return QtCore.QSize(self.width, self.height)

    def paintEvent(self, event):
        while self.node.retired_by:
            self.node = self.node.retired_by
        self.node.qpainter = QPainter()
        self.node.qpainter.begin(self)

        if not self.node.style_bgcolor is None:
            bgBrush = QtGui.QBrush()
            bgBrush.setColor(QtGui.QColor(self.node.style_bgcolor))
            bgBrush.setStyle(QtCore.Qt.SolidPattern)
            rect = QtCore.QRect(0, 0, self.node.qpainter.device().width, self.node.qpainter.device().height)
            self.node.qpainter.fillRect(rect, bgBrush)

        self.node.painter(self.node, *self.node.args)

        self.node.qpainter.end()
        self.node.qpainter = None

class Canvas(QtBaseWidget):
    def __init__(self, painter, *args):
        super().__init__()
        self.ui = None
        self.painter = painter
        self.args = args

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.ui.puinode = self
            self.ui.width = self.layout_width or 0
            self.ui.height = self.layout_height or 0
        else:
            self.ui = PUIQtCanvas(self, self.layout_width or 0, self.layout_height or 0)
        self.ui.update()
        super().update(prev)

    def drawText(self, x, y, text):
        self.qpainter.drawText(QPoint(int(x), int(y)), text)

    def drawLine(self, x1, y1, x2, y2, color=None, width=None):
        pen = self.qpainter.pen()
        orig_color = pen.color()
        orig_width = pen.width()
        if not color is None:
            pen.setColor(QColor(color))
        if not width is None:
            pen.setWidth(width)
        self.qpainter.setPen(pen)
        self.qpainter.drawLine(x1, y1, x2, y2)
        pen.setColor(orig_color)
        pen.setWidth(orig_width)
        self.qpainter.setPen(pen)

    def drawPolyline(self, coords, color=None, width=None):
        pen = self.qpainter.pen()
        orig_color = pen.color()
        orig_width = pen.width()
        if not color is None:
            pen.setColor(QColor(color))
        if not width is None:
            pen.setWidth(width)
        self.qpainter.setPen(pen)
        self.qpainter.drawPolyline([QtCore.QPointF(x,y) for x,y in coords])
        pen.setColor(orig_color)
        pen.setWidth(orig_width)
        self.qpainter.setPen(pen)



from .. import *
from .base import *

class QtTableModelAdapter(QtCore.QAbstractTableModel):
    def __init__(self, model):
        super().__init__()
        self.model = model

    def data(self, index, role):
        if role == QtCore.Qt.DisplayRole:
            return self.model.data(index.row(), index.column())
        elif role == QtCore.Qt.EditRole:
            if hasattr(self.model, "editData"):
                return self.model.editData(index.row(), index.column())
            else:
                return self.model.data(index.row(), index.column())

    def setData(self, index, value, role):
        if role == QtCore.Qt.EditRole:
            self.model.setData(index.row(), index.column(), value)
            return True

    def flags(self, index):
        flags = super().flags(index)
        if self.model.editable(index.row(), index.column()):
            flags |= QtCore.Qt.ItemIsEditable
        return flags

    def headerData(self, section, orientation, role):
        if role == QtCore.Qt.DisplayRole:
            if orientation == QtCore.Qt.Horizontal:
                if hasattr(self.model, "columnHeader"):
                    if self.model.columnHeader is None:
                        return None
                    else:
                        return self.model.columnHeader(section)
                else:
                    return super().headerData(section, orientation, role)
            else:
                if hasattr(self.model, "rowHeader"):
                    if self.model.rowHeader is None:
                        return None
                    else:
                        return self.model.rowHeader(section)
                else:
                    return super().headerData(section, orientation, role)

    def rowCount(self, index):
        return self.model.rowCount()

    def columnCount(self, index):
        return self.model.columnCount()


class Table(QtBaseWidget):
    def __init__(self, model, autofit=True):
        super().__init__()
        self.model = model
        self.autofit = autofit

    def update(self, prev):
        if prev and prev.ui and prev.qt_model:
            self.ui = prev.ui
            self.qt_model = prev.qt_model
        else:
            self.ui = QtWidgets.QTableView()
            self.qt_model = QtTableModelAdapter(self.model)

        if self.model.columnHeader is None:
            self.ui.horizontalHeader().hide()
        else:
            self.ui.horizontalHeader().show()

        if self.model.rowHeader is None:
            self.ui.verticalHeader().hide()
        else:
            self.ui.verticalHeader().show()

        self.ui.setModel(self.qt_model)
        if self.autofit:
            self.ui.resizeColumnsToContents()

        super().update(prev)



from .. import *
from .base import *

class Text(FBase):
    def __init__(self, text, selectable=False):
        super().__init__()
        self.text = text

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.ui.value = self.text
            try:
                self.ui.update()
            except:
                pass
        else:
            self.ui = ft.Text(self.text, expand=self.layout_weight)
        super().update(prev)

class Html(FBase):
    supported = False
    def __init__(self, text, **kwargs):
        super().__init__(**kwargs)
        self.text = text

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.ui.value = self.text
            try:
                self.ui.update()
            except:
                pass
        else:
            self.ui = ft.Text(self.text, expand=self.layout_weight)
        super().update(prev)

class MarkDown(FBase):
    supported = False
    def __init__(self, text, **kwargs):
        super().__init__(**kwargs)
        self.text = text

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.ui.value = self.text
            try:
                self.ui.update()
            except:
                pass
        else:
            self.ui = ft.Markdown(self.text, expand=self.layout_weight, auto_follow_links=True)
        super().update(prev)



from setuptools import setup, find_packages
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()
from PUI import __version__

setup(
    name='QPUIQ',
    version=__version__,
    url='https://github.com/buganini/PUI',
    author='Buganini Chiu',
    author_email='buganini@b612.tw',
    description='"PUI" Python Declarative UI Framework',
    packages=find_packages(),    
    long_description=long_description,
    long_description_content_type='text/markdown'
)



from .. import *
from .base import *
import flet.canvas as cv

class Canvas(FBase):
    def __init__(self, painter, *args):
        super().__init__()
        self.ui = None
        self.painter = painter
        self.args = args

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
            self.ui.puinode = self
        else:
            self.ui = cv.Canvas()
        if self.layout_width:
            self.ui.width = self.layout_width
        else:
            self.ui.width = float("inf")
        if self.layout_height:
            self.ui.height = self.layout_height
        else:
            self.ui.height = float("inf")
        self.ui.shapes.clear()
        self.painter(self, *self.args)
        try:
            self.ui.update()
        except:
            pass
        super().update(prev)

    def drawText(self, x, y, text):
        self.ui.shapes.append(
            cv.Text(
                x,
                y,
                text,
            )
        )

    def drawLine(self, x1, y1, x2, y2, color=None, width=None):
        params = {}
        if not color is None:
            params["color"] = f"#{color:06X}"
        if not width is None:
            params["stroke_width"] = width
        self.ui.shapes.append(
            cv.Line(
                x1,
                y1,
                x2,
                y2,
                ft.Paint(**params, style=ft.PaintingStyle.STROKE)
            ),
        )

    def drawPolyline(self, coords, color=None, width=None):
        params = {}
        if not color is None:
            params["color"] = f"#{color:06X}"
        if not width is None:
            params["stroke_width"] = width
        paint = ft.Paint(**params, style=ft.PaintingStyle.STROKE)
        for i in range(1, len(coords)):
            self.ui.shapes.append(
                cv.Line(
                    coords[i-1][0],
                    coords[i-1][1],
                    coords[i][0],
                    coords[i][1],
                    paint
                ),
            )



from .application import *
from .button import *
from .canvas import *
from .checkbox import *
from .label import *
from .layout import *
from .progressbar import *
from .radiobutton import *
from .scroll import *
from .tab import *
from .text import *
from .textfield import *
from .window import *

PUIView = TPUIView
class DummyWidget(PUINode):
    supported = False
    def __init__(self, *args, **kwargs):
        super().__init__()

Combobox = DummyWidget
ComboboxItem = DummyWidget

PUI_BACKEND = "tkinter"



from .. import *
from .base import *

class HBox(TkBaseWidget):
    use_ttk = "TFrame"
    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = ttk.Frame(self.tkparent.inner)
            self.ui.grid_rowconfigure(0, weight=1)
        super().update(prev)

    def putChild(self, idx, child):
        if isinstance(child, TkBaseWidget):
            child.outer.grid(row=0, column=idx, sticky='nsew')
            if child.layout_weight is None:
                self.ui.grid_columnconfigure(idx, weight=0)
            else:
                self.ui.grid_columnconfigure(idx, weight=child.layout_weight, uniform=".")
        elif child.children:
            self.putChild(idx, child.children[0])

    def removeChild(self, idx, child):
        self.ui.grid_columnconfigure(idx, weight=0)
        if isinstance(child, TkBaseWidget):
            child_outer = child.outer
            if child_outer:
                child_outer.grid_forget()
        elif child.children:
            self.removeChild(idx, child.children[0])

    def postSync(self):
        for i,child in enumerate(self.children):
            self.putChild(i, child)
        super().postSync()

class VBox(TkBaseWidget):
    use_ttk = "TFrame"
    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = ttk.Frame(self.tkparent.inner)
            self.ui.grid_columnconfigure(0, weight=1)
        super().update(prev)

    def putChild(self, idx, child):
        if isinstance(child, TkBaseWidget):
            child.outer.grid(row=idx, column=0, sticky='nsew')
            if child.layout_weight is None:
                self.ui.grid_rowconfigure(idx, weight=0)
            else:
                self.ui.grid_rowconfigure(idx, weight=child.layout_weight, uniform=".")
        elif child.children:
            self.putChild(idx, child.children[0])

    def removeChild(self, idx, child):
        self.ui.grid_rowconfigure(idx, weight=0)
        if isinstance(child, TkBaseWidget):
            child_outer = child.outer
            if child_outer:
                child_outer.grid_forget()
        elif child.children:
            self.removeChild(idx, child.children[0])

    def postSync(self):
        for i,child in enumerate(self.children):
            self.putChild(i, child)
        super().postSync()

class Spacer(TkBaseWidget):
    use_ttk = "TFrame"
    def __init__(self, *args):
        super().__init__(*args)
        self.layout_weight = 1

    def update(self, prev):
        if prev and prev.ui:
            self.ui = prev.ui
        else:
            self.ui = ttk.Frame(self.tkparent.inner)
